"""
BBOP Graph class created in the original biolink-api
before ontobio was stripped out, it is still used
in the scigraph-util

TODO Merge this with OBOGraph
"""
from typing import Dict


class BBOPGraph:

    """
    BBOPGraph Graph object model
    https://github.com/berkeleybop/bbop-graph
    """

    nodemap = {}

    def __init__(self, obj: Dict = None):
        obj = obj or {}
        self.nodes = []
        self.edges = []
        if obj:
            self.add_json_graph(obj)

    def add_json_graph(self, obj):
        for node in obj['nodes']:
            self.add_node(Node(**node))
        for edge in obj['edges']:
            self.add_edge(Edge(edge))

    def add_node(self, node) :
        self.nodemap[node.id] = node
        self.nodes.append(node)

    def add_edge(self, edge) :
        self.edges.append(edge)

    def merge(self, graph):
        for node in graph.nodes:
            self.add_node(node)
        for edge in graph.edges:
            self.add_edge(edge)

    def get_node(self, id):
        return self.nodemap[id]

    def get_lbl(self, id):
        return self.nodemap[id].lbl

    def get_root_nodes(self, relations):
        roots = []
        if relations is None:
            relations = []
        for node in self.nodes:
            if len(self.get_outgoing_edges(node.id, relations)) == 0:
                roots.append(node)
        return roots

    def get_leaf_nodes(self, relations):
        roots = []
        if relations is None:
            relations = []
        for node in self.nodes:
            if len(self.get_incoming_edges(node.id, relations)) == 0:
                roots.append(node)
        return roots

    def get_outgoing_edges(self, nid, relations):
        el = []
        if relations is None:
            relations = []
        for edge in self.edges:
            if edge.sub == nid:
                if len(relations) == 0 or edge.pred in relations:
                    el.append(edge)
        return el

    def get_incoming_edges(self, nid, relations=[]):
        el = []
        for edge in self.edges:
            if edge.obj == nid:
                if len(relations) == 0 or edge.pred in relations:
                    el.append(edge)
        return el

    def as_dict(self):
        return {
            "nodes": [node.as_dict() for node in self.nodes],
            "edges": self.edges
        }


class Node:
    def __init__(self, id, lbl=None, meta=None):    
        self.id = id
        self.lbl = lbl
        self.meta = Meta(meta)

    def __str__(self):
        return self.id + ' "' + str(self.lbl) + '"'

    def as_dict(self):
        return {
            "id": self.id,
            "lbl": self.lbl,
            "meta": self.meta.pmap
        }


class Edge:
    def __init__(self, obj):
        self.sub = obj['sub']
        self.pred = obj['pred']
        self.obj = obj['obj']
        self.meta = obj['meta']
    
    def __str__(self):
        return self.sub + "-[" + self.pred + "]->" + self.obj


class Meta:
    def __init__(self, obj):
        self.type_list = obj['types']
        self.category_list = []
        if 'category' in obj:
            self.category_list = obj['category']
        self.pmap = obj



from dataclasses import dataclass
from typing import Dict, List, Optional


@dataclass(frozen=True)
class Highlight:
    highlight: str = None
    match: str = None
    has_highlight: bool = None


@dataclass
class SearchResults:
    """
    Search results class from transformed solr results

    docs: pysolr.Results.docs, docs returned by solr (schema dependent)

    numFound: count of total docs for query,
              {HTTPResponse.json()}.response.numFound or pysolr.Results.docs

    facet_count:
        A solr facet field looks like this: [field1, count1, field2, count2, ..., fieldN, countN]

          We translate this to a dict {f1: c1, ..., fn: cn}

    highlighting: dictionary of highlights per id:
        ex: {id1: "<em> some highlight </em> some not highlighted"}

    """
    docs: List[Dict]
    numFound: int
    facet_counts: Dict
    highlighting: Optional[Dict[str, str]] = None


@dataclass
class AutocompleteResult:
    """
    Data class for single autocomplete result

    id: curie formatted id
    label: primary label (rdfs:label)
    match: matched part of document (may be primary label, synonym, id, etc)
    category: Neo4J label(s)
    taxon: taxon as NCBITaxon curie
    taxon_label: taxon label
    highlight: solr highlight
    has_highlight: True if highlight can be interpreted as html, else False
    """

    id: str
    label: str
    match: str
    category: List[str]
    taxon: str
    taxon_label: str
    highlight: str
    has_highlight: bool
    equivalent_ids: List[str]



"""
Adapter between MME API model and Monarch sim API

See https://github.com/ga4gh/mme-apis/blob/87f615/search-api.md

Original API code here: https://github.com/monarch-initiative/monarch-mme
"""
from typing import List

from ontobio.model.mme.request import MmeRequest, Observed
from ontobio.model.mme.response import MmeResponse, Result, Score
from ontobio.golr.golr_query import GolrSearchQuery
from ontobio.sim.phenosim_engine import PhenoSimEngine


def clean_feature_ids(id: str) -> str:
    """
    MME queries often need to be sanitized before going into owlsim, for example:
    MIM:610536 -> OMIM:610536
    SHH -> HGNC:10848
    """
    if ':' in id:
        prefix, reference = id.split(':')
        if prefix == 'MIM':
            id = 'OMIM:' + reference
    else:
        # Assume it's a label and look it up
        # Assume it's human, and make sure it's an exact match
        query = GolrSearchQuery(id, taxon=['NCBITaxon:9606'], min_match="100%")
        results = query.search()
        if results.docs:
            id = results.docs[0]['id']

    return id


def extract_features_from_mme(patient: MmeRequest) -> List:
    """
    Extract features or diseases and genes of interest from MME request object

    If the record contains feature(s) with is_observed=True will we will send
    those to owlsim, otherwise we will the disease(s) of interest, if there
    are no diseases, will send and gene(s) of interest to owlsim.  Essentially
    the goal is to send the smallest profile to owlsim to get the best match
    """
    features = [clean_feature_ids(feature.id) for feature in patient.features if feature.observed == Observed.yes]
    diseases = [clean_feature_ids(disease.id) for disease in patient.disorders]
    genes = [clean_feature_ids(genomic_feature.gene.id) for genomic_feature in patient.genomicFeatures]
    if features:
        return features
    elif diseases:
        return diseases
    elif genes:
        return genes


def query_owlsim(patient: MmeRequest, sim_engine: PhenoSimEngine, taxon: str = None) -> MmeResponse:
    """
    Query owlsim with features from MmeRequest object,
    see extract_features_from_mme for what features are selected to
    be sent to owlsim
    """
    features = extract_features_from_mme(patient)
    results = sim_engine.search(
            id_list=features,
            limit=100,
            taxon_filter=taxon,
            is_feature_set=False
        )
    mme_response = MmeResponse()

    for result in results.matches:
        mme_score = Score(patient=result.score)
        mme_result = Result(mme_score)
        if result.id.startswith('MONDO'):
            mme_result._disease = result.id
            mme_result._disease_label = result.label
        else:
            mme_result._gene = result.id
            mme_result._gene_label = result.label

        mme_response.results.append(mme_result)

    return mme_response



"""
MME request object, see
https://github.com/ga4gh/mme-apis/blob/87f615/search-api.md
"""

from dataclasses import dataclass, field
from enum import Enum
from typing import List


class Sex(str, Enum):
    FEMALE = 'FEMALE'
    MALE = 'MALE'
    OTHER = 'OTHER'
    MIXED_SAMPLE = 'MIXED_SAMPLE'
    NOT_APPLICABLE = 'NOT_APPLICABLE'


class Observed(str, Enum):
    yes = 'yes'
    no = 'no'

@dataclass
class Contact:
    name: str = None
    institution:  str = None
    href:  str = None
    email: str = None
    roles: List[str] = field(default_factory=list)

@dataclass
class Patient:
    id: str
    label: str = None
    contact: Contact = None

@dataclass
class Disorder:
    id: str
    label: str = None


@dataclass
class Feature:
    id: str
    label: str = None
    observed: Observed = Observed.yes
    ageOfOnset: str = None


@dataclass
class Gene:
    id: str


@dataclass
class Variant:
    assembly: str
    referenceName: str
    start: int
    end: int = None
    referenceBases: str = None
    alternateBases: str = None


@dataclass
class GenomicFeatureType:
    id: str
    label: str = None


@dataclass
class GenomicFeature:
    gene: Gene
    variant: Variant = None
    zygosity: int = None
    type: GenomicFeatureType = None


@dataclass
class MmeRequest:
    id: str
    label: str = None
    disclaimer: str = None
    contact: Contact = None
    terms: str = None
    patient: Patient = None
    species: str = None
    sex: Sex = None,
    ageOfOnset: str = None,
    inheritanceMode: str = None,
    disorders: List[Disorder] = field(default_factory=list)
    features: List[Feature] = field(default_factory=list)
    genomicFeatures: List[GenomicFeature] = field(default_factory=list)
    test: bool = False

    def __post_init__(self):
        if not self.features and not self.genomicFeatures:
            raise ValueError("Request must container either features or genomicFeatures")
